#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ranlib.h"
#include "util.h"
#include "blackScholes.h"
#include "european.h"

#define SMALL 0.0001

double Mean,
  Variance,
  Dividend,
  Initial,
  Strike,
  Interest,
  Expiration,
  A0,
  A1,
  A2,
  *Parm,
  *Cov,
  *Scratch,
  **Next_price,
  **Price,
  **Price_swap,
  *Next_vih,
  *Vih,
  *Vih_swap,
  **Next_index,
  **Index,
  **Index_swap;



int Timeslices,
  Underliers,
  Meshpoints,
  Next_Meshpoints,
  Repetitions,
  Effective_underliers;

void init_rng(int time);
double weight(int i, int j, int time);
double mesh_estimate0(int i, int t);
double mesh_estimate1(int i, int t);
double mesh_estimate3(int i, int t);
double mesh_estimate4(int i, int t);
void gen_index(int dim, double *value, double *index);

int main(int ARGC, char *ARGV[]){
  long seed1, seed2;
  int i,j,t,max_element,rep;
  double w,sumw, *growth,f,g,G,beta;
  double S1,S2,S7,S8,S9,S10,S11,S12,S13,S14,S15,S16,S17;
  time_t start,end;

  if(ARGC != 16){
    printf("%s Timeslices Expiration Variance Dividend Interest Initial Strike Underliers A0 A1 A2 Repetitions Seed1 Seed2 Effective_underliers\n",ARGV[0]);
    printf("Wrong number of arguments\n");
    return(1);
  }

  sscanf(ARGV[1],"%i",&Timeslices);
  sscanf(ARGV[2],"%lf",&Expiration);
  sscanf(ARGV[3],"%lf",&Variance); Variance = Variance * Expiration /(double)(Timeslices-1);
  sscanf(ARGV[4],"%lf",&Dividend); Dividend = Dividend * Expiration /(double)(Timeslices-1);
  sscanf(ARGV[5],"%lf",&Interest); Interest = Interest * Expiration /(double)(Timeslices-1);
  sscanf(ARGV[6],"%lf",&Initial);
  sscanf(ARGV[7],"%lf",&Strike);
  sscanf(ARGV[8],"%i",&Underliers);
  sscanf(ARGV[9],"%lf",&A0);
  sscanf(ARGV[10],"%lf",&A1);
  sscanf(ARGV[11],"%lf",&A2);
  sscanf(ARGV[12],"%i",&Repetitions);
  sscanf(ARGV[13],"%i",&seed1);
  sscanf(ARGV[14],"%i",&seed2);
  sscanf(ARGV[15],"%i",&Effective_underliers);

  printf("Timeslices: 0 - %i\n",Timeslices-1);
  printf("Expiration: %lf\n",Expiration);
  printf("Variance:   %lf per timeslice, %lf per period\n",Variance, (double)(Timeslices-1)*Variance/Expiration);
  printf("Dividend:   %lf per timeslice, %lf per period\n",Dividend, (double)(Timeslices-1)*Dividend/Expiration);
  printf("Interest:   %lf per timeslice, %lf per period\n",Interest, (double)(Timeslices-1)*Interest/Expiration);
  printf("Initial:    %lf\n",Initial);
  printf("Strike:     %lf\n",Strike);
  printf("Underliers: %i\n",Underliers);
  printf("Meshpoints: %lf + %lf*t + %lf*t*t\n",A0,A1,A2);
  printf("Repetitons: %i\n",Repetitions);

  Meshpoints = (int)(A0 + A1*(Timeslices-1) + A2*pow((Timeslices-1),(double)(Underliers-1)/2.0));

  printf("Last timeslice points: %i\n",Meshpoints);

  allocate2dDouble(&Next_price, Meshpoints, Underliers);
  allocate2dDouble(&Price, Meshpoints, Underliers);
  allocate1dDouble(&Next_vih, Meshpoints);
  allocate1dDouble(&Vih, Meshpoints);
  allocate2dDouble(&Next_index, Meshpoints, Underliers);  
  allocate2dDouble(&Index, Meshpoints, Underliers);
  allocate1dDouble(&Cov,Underliers*Underliers);
  allocate1dDouble(&Parm,Underliers*(Underliers+3)/2+1);
  allocate1dDouble(&growth, Underliers);
  allocate1dDouble(&Scratch,Underliers);

  if(seed1 == seed2 == 0){
    seed1 = time(NULL);
    srand(seed1);
    seed2 = rand();
  }
  setall(seed1, seed2);

  if(Effective_underliers == 0) Effective_underliers = Underliers;

  for(rep=1;rep<=Repetitions;++rep){
    start = time(NULL);
    t = Timeslices -1;
    Meshpoints = (int)(A0 + A1*t + A2*pow(t,(double)(Underliers-1)/2.0));
    init_rng(t);
    for(i=0; i<Meshpoints; i++){
      genmn(Parm, growth, Scratch);
      for(j=0; j<Underliers; j++){
	Price[i][j] = Initial*exp(growth[j] - t*Variance/2.0);
      }
      gen_index(Underliers,Price[i],Index[i]);
      Vih[i] = 0;
    }
    
    for(t = Timeslices-2; t>0; t--){
      Next_Meshpoints = Meshpoints; 
      Meshpoints = (int)(A0 + A1*t + A2*pow(t,(double)(Underliers-1)/2.0));

      Price_swap = Price;
      Price = Next_price;
      Next_price = Price_swap;

      Vih_swap = Vih;
      Vih = Next_vih;
      Next_vih = Vih_swap;

      Index_swap = Index;
      Index = Next_index;
      Next_index = Index_swap;
  
      init_rng(t);
      for(i=0; i<Meshpoints; i++){
	genmn(Parm, growth, Scratch);
	for(j=0; j<Underliers; j++){
	  Price[i][j] = Initial*exp(growth[j] - t*Variance/2.0);
	}
	gen_index(Underliers, Price[i], Index[i]);
	Vih[i] = mesh_estimate3(i,t);
      }
    }
    
    Price_swap = Price;
    Price = Next_price;
    Next_price = Price_swap;

    Vih_swap = Vih;
    Vih = Next_vih;
    Next_vih = Vih_swap;

    Index_swap = Index;
    Index = Next_index;
    Next_index = Index_swap;
        
    for(j=0; j<Underliers; j++){
      Price[0][j] = Initial;
    }

    Next_Meshpoints = Meshpoints;
    Vih[0] = mesh_estimate3(0,0);
    end = time(NULL);
    printf("%lf\t%i\n",Vih[0],(int)(end - start));
  }
}

double weight(int i, int j, int time){
  double arg1 = 0, arg2 = 0, growth;
  int u;

  for(u=0; u<Underliers; u++){
    if(Index[j][u] < Effective_underliers){
      growth = log(Next_price[j][u]/Price[i][u]) + Variance/2.0;
      arg1 += (growth - (Interest-Dividend))*(growth - (Interest-Dividend))/Variance;

      growth = log(Next_price[j][u]/Initial) + (time+1)*Variance/2.0;
      arg2 += (growth - (time+1)*(Interest-Dividend))*(growth - (time+1)*(Interest-Dividend))/((time+1)*Variance);
    }
  }

  /*exp(-0.5*arg1) / exp(-0.5*arg2)*/
  return(exp(-0.5*(arg1 - arg2)));
}

void init_rng(int time){
  int i,j,k;
  double *dam;
                                 
  allocate1dDouble(&dam, Underliers);
  k = 0;
  for(i=0; i<Underliers; i++){
    for(j=0; j<Underliers; j++){
      if(i == j){
	Cov[k] = Variance*time;
      }else{
	Cov[k] = 0.0;
      }
      k++;
    }
  }

  for(i=0; i<Underliers; i++){
    dam[i] = (Interest - Dividend)*time;
  }

  setgmn(dam, Cov, Underliers, Parm);
  deallocate1dDouble(&dam, Underliers);
}


double mesh_estimate0(int i, int t){
  double Vih =0, sumw = 0, max,w;
  int j;

  for(j=0; j<Meshpoints; j++){
    max = dArrayMax(Underliers, Next_price[j]);
    w = weight(i,j,t);
    sumw += w;
    if(Next_vih[j] > dMax(0,max - Strike)){
      Vih += w*Next_vih[j];
    }else{
      Vih += w*exp(-Interest*(t+1))*dMax(0,max - Strike);
    }
  }
  Vih /= sumw;
  return(Vih);
}    

double mesh_estimate1(int i, int t){
  double S1,S2,S7,S8,S9,S10,S12,S13,w,f,g,G,beta,max;
  int u,j, max_element;
      
  max_element = dWhichMax(Underliers, Price[i]);
  
  G = exp(-Interest*t)*euroCall(Price[i][max_element], Strike, Interest, sqrt(Variance), 1.0 , Dividend);  

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S12 = 0;
  S13 = 0;

  for(j=0; j<Meshpoints; j++){
    w = weight(i,j,t);
    g = dMax(0, exp(-Interest*(t+1))*(Next_price[j][max_element] - Strike));
    max = dArrayMax(Underliers, Next_price[j]);
    f = dMax(0, exp(-Interest*(t+1))*(max - Strike));
    f = dMax(f, Next_vih[j]);
	     
    S1 += w;
    S2 += w*w;
    S7 += w*f;
    S8 += w*(g-G);
    S9 += w*w*f;
    S10 += w*w*(g-G);
    S12 += w*w*f*(g-G);
    S13 += w*w*(g-G)*(g-G);    
  }

  if( (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1)) == 0){
    return(S7/S1);
  }

  beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/
    (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));

  return((S7-beta*S8)/S1);
}


double mesh_estimate3(int i, int t){
  double S1,S2,S7,S8,S9,S10,S12,S13,w,f,g,G,beta,max;
  int u,j, max1,max2,wmax,k,choice,exercise;
      
  if(Underliers == 1) return(mesh_estimate1(i,t));

  if(Price[i][0] > Price[i][1]){
    max1 = 0;
    max2 = 1;
  }else{
    max1 = 1;
    max2 = 0;
  }

  for(u=2; u<Underliers; ++u){
    if(Price[i][u] > Price[i][max1]){
      max2 = max1;
      max1 = u;
    }else if(Price[i][u] > Price[i][max2]){
      max2 = u;
    }
  }
  
  G = exp(-Interest*t)*maxOf2Risk(Price[i][max1], Price[i][max2], Strike, 1.0, sqrt(Variance), sqrt(Variance),
				  Interest, Dividend, Dividend, 0.0);

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S12 = 0;
  S13 = 0;

   for(j=0; j<Next_Meshpoints; j++){ 
/*   for(j=0; j<Meshpoints; j++){ */
    w = weight(i,j,t);

    g = dMax(Next_price[j][max1],Next_price[j][max2]);
    g = dMax(0, exp(-Interest*(t+1))*(g - Strike));

    max = dArrayMax(Underliers, Next_price[j]);

    f = dMax(0, exp(-Interest*(t+1))*(max - Strike));
    if(f < Next_vih[j]){
      f = Next_vih[j];
      exercise = 0;
    }else{
      exercise = 1;
    }

    wmax = dWhichMax(Underliers, Next_price[j]);
/*     choice = 1; */
/*     for(k=0;k<Underliers;k++){ */
/*       if(Price[i][wmax] < Price[i][k]){ */
/* 	++choice; */
/*       } */
/*     } */

    /* don't add points which would exercise an underlier which is not an effective underlier */
    if(!(Index[j][wmax] >= Effective_underliers && exercise)){
      S1 += w;
      S2 += w*w;
      S7 += w*f;
      S8 += w*(g-G);
      S9 += w*w*f;
      S10 += w*w*(g-G);
      S12 += w*w*f*(g-G);
      S13 += w*w*(g-G)*(g-G);    
    }
  }

   if( (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1)) == 0){
     return(S7/S1);
   }

  beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/
    (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));

  if(fabs(beta) > 2.0){
     return(S7/S1);
  }    

/*   printf("%12lf - %12lf*%12lf = %12lf - %12lf\tRatio = %12lf\n",S7/S1,beta,S8/S1,S7/S1,beta*S8/S1,(S7-beta*S8)/S1, fabs(beta*S8/S7)); */

  return((S7-beta*S8)/S1);
}

double mesh_estimate4(int i, int t){
  double S1,S2,S7,S8,S9,S10,S12,S13,w,f,g,G,beta,max;
  int u,j, max_element;
      
  max_element = dWhichMax(Underliers, Price[i]);
  
  G = exp(-Interest*t)*european(Underliers,Price[i], Strike, Interest, Variance, Dividend, 0.01);  

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S12 = 0;
  S13 = 0;

  for(j=0; j<Meshpoints; j++){
    w = weight(i,j,t);
    max = dArrayMax(Underliers, Next_price[j]);
    g = dMax(0, exp(-Interest*(t+1))*(max - Strike));
    f = dMax(g, Next_vih[j]);
	     
    S1 += w;
    S2 += w*w;
    S7 += w*f;
    S8 += w*(g-G);
    S9 += w*w*f;
    S10 += w*w*(g-G);
    S12 += w*w*f*(g-G);
    S13 += w*w*(g-G)*(g-G);    
  }

  beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/
    (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));

  return((S7-beta*S8)/S1);
}

/* Slow, implement quick or heap sort version later */
void gen_index(int dim, double *value, double *index){
  int i,j;

  for(i=0;i<dim;++i) index[i] = dim;

  for(i=0;i<dim;++i){
    for(j=0;j<i;++j){
      if(value[i] > value[j]){
	--index[i];
      }else if(value[j] > value[i]){
	--index[j];
      }
    }
  }
}
